/* 
 * Sound Monitor applet for the GNOME panel
 * (C) 2002 John Ellis
 *
 * Author: John Ellis
 *
 * This software is released under the GNU General Public License (GNU GPL).
 * Please read the included file COPYING for more information.
 * This software comes with no warranty of any kind, use at your own risk!
 */

#include "sound-monitor.h"
#include "skin.h"
#include "update.h"
#include "ui_pixbuf_ops.h"
#include "ui2_main.h"
#include "ui2_skin.h"
#include "ui2_parse.h"

#include "ui2_item.h"
#include "ui2_meter.h"
#include "ui2_scope.h"
#include "ui2_spectrum.h"

/* the default pixmaps */
#include "def_theme/back_h.xpm"
#include "def_theme/led_h.xpm"
#include "def_theme/status_h.xpm"
#include "def_theme/back_v.xpm"
#include "def_theme/led_v.xpm"
#include "def_theme/status_v.xpm"


SkinData *applet_skin_default(UIData *ui, const gchar *key, gpointer data)
{
	SkinData *skin;

	skin = skin_new();

	if (key && (strcmp(key, "skindata_h_tiny") == 0 ||
		    strcmp(key, "skindata_h_small") == 0 ||
		    strcmp(key, "skindata_v") == 0 ||
		    strcmp(key, "skindata_v_large") == 0 ||
		    strcmp(key, "skindata_v_huge") == 0 ))
		{
		skin->real_overlay = gdk_pixbuf_new_from_xpm_data((const char **)back_h_xpm);

		meter_register_to_skin("left", (gchar **)led_h_xpm, 5, 1,
				       14, TRUE, FALSE, skin, NULL);
		meter_register_to_skin("right", (gchar **)led_h_xpm, 5, 10,
				       14, TRUE, FALSE, skin, NULL);

		item_register_to_skin("status", (gchar **)status_v_xpm, 1, 1,
				      4, skin, NULL);
		}
	else
		{
		skin->real_overlay = gdk_pixbuf_new_from_xpm_data((const char **)back_v_xpm);

		meter_register_to_skin("left", (gchar **)led_v_xpm, 1, 1,
				       14, FALSE, FALSE, skin, NULL);
		meter_register_to_skin("right", (gchar **)led_v_xpm, 10, 1,
				       14, FALSE, FALSE, skin, NULL);

		item_register_to_skin("status", (gchar **)status_h_xpm, 1, 45,
				      4, skin, NULL);
		}

	skin->width = gdk_pixbuf_get_width(skin->real_overlay);
	skin->height = gdk_pixbuf_get_height(skin->real_overlay);

	skin->width_def = skin->width_min = skin->width_max = skin->width;
	skin->height_def = skin->height_min = skin->height_max = skin->height;

	return skin;
}

static gint applet_skin_backing_sync_idle(gpointer data)
{
	AppData *ad = data;

	applet_skin_backing_sync(ad);

	return FALSE;
}

static void applet_skin_back_rgb_free(guchar *pixels, gpointer data)
{
	g_free(pixels);
}

void applet_skin_backing_sync(AppData *ad)
{
	guchar *rgb = NULL;
	int w, h, rs;

	if (!ad->ui || !ad->ui->skin || !ad->ui->skin->pixbuf) return;

	if (ad->ui->skin->transparent)
		{
		skin_sync_back(ad->ui->skin, ad->ui);
		return;
		}

	if (!gdk_pixbuf_get_has_alpha(ad->ui->skin->real_overlay)) return;

	applet_widget_get_rgb_bg(APPLET_WIDGET(ad->applet),
				 &rgb, &w, &h, &rs);

	if (rgb == NULL) return;

	if (w != gdk_pixbuf_get_width(ad->ui->skin->pixbuf) || h != gdk_pixbuf_get_height(ad->ui->skin->pixbuf))
		{
		/* out of sync, this happens in startup, when changing
		 * panel size, switching panels, etc. So fill the back with the first color, may be
		 * ugly on pixmap backs, but better than showing unzeroed memory till the next sync.
		 */

		GdkPixbuf *pb;
		GdkPixbuf *new_pb;
		guchar *p = rgb;
		gint r, g, b;
		gint pw, ph;

		r = *p; p++;
		g = *p; p++;
		b = *p;

		pw = gdk_pixbuf_get_width(ad->ui->skin->pixbuf);
		ph = gdk_pixbuf_get_height(ad->ui->skin->pixbuf);

		pb = gdk_pixbuf_new(GDK_COLORSPACE_RGB, FALSE, 8, pw, ph);
		pixbuf_draw_rect_fill(pb, 0, 0, pw, ph, r, g, b, 255);

		/* set what we do have */
		w = MIN(w, pw);
		h = MIN(h, ph);
		new_pb = gdk_pixbuf_new_from_data(rgb, GDK_COLORSPACE_RGB, FALSE, 8,
							  w, h, rs, applet_skin_back_rgb_free, NULL);
		rgb = NULL;
		pixbuf_copy_area(new_pb, 0, 0, pb, 0, 0, w, h, FALSE);
		gdk_pixbuf_unref(new_pb);

		ui_set_underlay(ad->ui, pb);
		gdk_pixbuf_unref(pb);

		/* bah, newer panels don't seem to send the do_draw signal once the applet size is synced.
		   probably because it thinks we already got the correct size here. (wrong!)
		   so, add an idle callback to do this all over again *sigh*
		 */
		g_idle_add(applet_skin_backing_sync_idle, ad);
		}
	else
		{
		GdkPixbuf *pb;

		pb = gdk_pixbuf_new_from_data(rgb, GDK_COLORSPACE_RGB, FALSE, 8,
					      w, h, rs, applet_skin_back_rgb_free, NULL);
		rgb = NULL;
		ui_set_underlay(ad->ui, pb);
		gdk_pixbuf_unref(pb);
		}
}

gint applet_skin_back_set_cb(UIData *ui, GdkPixbuf *pb, gpointer data)
{
	AppData *ad = data;
	GdkPixbuf *new_pb;
	guchar *rgb = NULL;
	int w, h, rs;
	gint pw, ph;

	applet_widget_get_rgb_bg(APPLET_WIDGET(ad->applet),
				 &rgb, &w, &h, &rs);

	if (rgb == NULL) return FALSE;

	new_pb = gdk_pixbuf_new_from_data(rgb, GDK_COLORSPACE_RGB, FALSE, 8,
					  w, h, rs, applet_skin_back_rgb_free, NULL);
	rgb = NULL;

	pw = gdk_pixbuf_get_width(pb);
	ph = gdk_pixbuf_get_height(pb);

	if (pw != w || pw != ph)
		{

		/* if larger must resync when idle :( */
		if (w < pw || h < ph)
			{
			g_idle_add(applet_skin_backing_sync_idle, ad);
			}

		w = MIN(w, gdk_pixbuf_get_width(pb));
		h = MIN(h, gdk_pixbuf_get_height(pb));
		}

	pixbuf_copy_area(new_pb, 0, 0, pb, 0, 0, w, h, FALSE);
	gdk_pixbuf_unref(new_pb);

	return TRUE;
}

static gchar *skin_mode_key(gint vertical, gint size)
{
	const gchar *stext;

	switch (size)
		{
		case SIZEHINT_TINY:
			stext = "_tiny";
			break;
		case SIZEHINT_SMALL:
			stext = "_small";
			break;
		case SIZEHINT_LARGE:
			stext = "_large";
			break;
		case SIZEHINT_HUGE:
			stext = "_huge";
			break;
		case SIZEHINT_STANDARD:
		default:
			stext = "";
			break;
		}

	return g_strdup_printf("skindata%s%s", (vertical) ? "_v" : "_h", stext);
}

static SkinData *skin_load(const gchar *skin_path, gint vertical, gint size, gint edit, gchar **key)
{
	SkinData *skin;
	gchar *buf;

	*key = skin_mode_key(vertical, size);
	buf = g_strconcat(skin_path, "/", *key, NULL);
	skin = skin_parse(skin_path, buf, edit);
	g_free(buf);

	if (!skin)
		{
		g_free(*key);
		*key = NULL;
		}

	return skin;
}

static SkinData *skin_load_best_match(const gchar *path, gint vertical, gint size, gint edit, gchar **key)
{
	SkinData *s = NULL;

	s = skin_load(path, vertical, size, edit, key);
	if (s) return s;

	/* tiny falls back to small */
	if (size == SIZEHINT_TINY)
		{
		s = skin_load(path, vertical, SIZEHINT_SMALL, edit, key);
		if (s) return s;
		}

	/* small falls back to tiny */
	if (size == SIZEHINT_SMALL)
		{
		s = skin_load(path, vertical, SIZEHINT_TINY, edit, key);
		if (s) return s;
		}

	/* huge falls back to large */
	if (size == SIZEHINT_HUGE)
		{
		s = skin_load(path, vertical, SIZEHINT_LARGE, edit, key);
		if (s) return s;
		}

	if (size == SIZEHINT_STANDARD) return NULL;

	/* failed so far, try standard */
	return skin_load(path, vertical, SIZEHINT_STANDARD, edit, key);
}

gint skin_set(const gchar *path, AppData *ad)
{
	SkinData *new_s = NULL;
	gchar *key = NULL;

	if (!path)
		{
		key = skin_mode_key((ad->orient == ORIENT_LEFT || ad->orient == ORIENT_RIGHT), ad->sizehint);
		new_s = ui_skin_load_default(ad->ui, key);
		}
	else
		{
		new_s = skin_load_best_match(path,
					     (ad->orient == ORIENT_LEFT || ad->orient == ORIENT_RIGHT),
					     ad->sizehint, FALSE, &key);
		if (!new_s)
			{
			new_s = skin_load(path,
					  !(ad->orient == ORIENT_LEFT || ad->orient == ORIENT_RIGHT),
					  SIZEHINT_STANDARD, FALSE, &key);
			}
		}

	if (!new_s)
		{
		g_free(key);
		return FALSE;	
		}

	ui_skin_set(ad->ui, new_s, path, key);
	applet_skin_backing_sync(ad);

	ad->has_meter = (ui_widget_exists(ad->ui, "main", meter_type_id()) ||
			 ui_widget_exists(ad->ui, "left", meter_type_id()) ||
			 ui_widget_exists(ad->ui, "right", meter_type_id()) ||
			 ui_widget_exists(ad->ui, "main", item_type_id()) ||
			 ui_widget_exists(ad->ui, "left", item_type_id()) ||
			 ui_widget_exists(ad->ui, "right", item_type_id()) );

	ad->has_scope = ui_widget_exists(ad->ui, "main", scope_type_id());

	ad->has_spectrum = (ui_widget_exists(ad->ui, "main", spectrum_type_id()) ||
			    ui_widget_exists(ad->ui, "left", spectrum_type_id()) ||
			    ui_widget_exists(ad->ui, "right", spectrum_type_id()) );


	g_free(key);

	return TRUE;
}

